 /* Test program for PID Controller
*/
#include <Wire.h>
#include <Adafruit_PWMServoDriver.h>
#include "clsPCA9555.h"
#include "SPI.h" // Include the Arduino SPI library

// Change X to the address you want
#define BOARD_ADDRESS "OpenPneu2.0"
#define PI_CONST 3.1415926535897932384626433832795

Adafruit_PWMServoDriver pwm0 = Adafruit_PWMServoDriver(0x48);
Adafruit_PWMServoDriver pwm1 = Adafruit_PWMServoDriver(0x44);


String inputString = "";         // a string to hold incoming data
int DAC1_Value = 0;
int DAC2_Value = 0;
volatile byte ABP_data[2];

// Define pins of ioport_1 (for sensor)
PCA9555 ioport_1 (0x20); // for sensor and DAC
#define S00 0
#define S01 1
#define S02 2
#define S03 3
#define S04 4
#define S05 5
#define S06 6
#define S07 7
#define S08 8
#define S09 9
int SSPIN[10] = {0,1,2,3,4,5,6,7,8,9};

//volatile float pressureKPa = 0.0;

/*
  pwm generator pin definition
  all indices from 0
*/

//In each air channel, the pwm Pin that controls the inflation pump flow rate
//e.g.,0th air channel's inflation pump 
//	 is controlled by AirChannelInflationPumpPwmPort[0] pin of pwm generator chip AirChannelInflationPumpPwmChip[0].
int AirChannelInflationPumpPwmPort[10] = {0,1,2,3,4,5,6,7,8,9};
int AirChannelInflationPumpPwmChip[10] = {0,0,0,0,0,0,0,0,0,0};

int AirChannelDeflationPumpPwmPort[10] = {10,11,12,13,14,15,  0,1,2,3};
int AirChannelDeflationPumpPwmChip[10] = {0,0,0,0,0,0,  1,1,1,1};

int AirChannelValvePwmPort[10] = {4,5,6,7,8,9,  10,11,12,13};
int AirChannelValvePwmChip[10] = {1,1,1,1,1,1,  1,1,1,1};

//valve state: when it is false, meaning that this valve is around 0.0 volt.
//              when it is true, meaning that this valve is around 3.3 volt.
                               //C0     C1      C2     C3    C4    
volatile bool valveState[10] = {false, false, false, false, false, 
                                false, false, false, false, false};
                               //C5     C6      C7     C8    C9 
/*PID difinition*/   
//When target pressure is positive:
//  1. when target pressure is higher than current one, inflation

//  2. when target pressure is lower than current one (larger than 2 KPa), only switch the air channel to deflation until pressure 
//      is below target pressure, then turn to situation 1

//When target pressure is negative:
//  1. when target pressure is lower than current one, deflation
//  2. when target pressure is higher than current one (larger than 2 KPa), only switch the air channel to inflation until pressure
//      is beyond target pressure, then turn to situation 1


float thresholdP = 2.0;
float KP_Original[10] = {9000.0, 9000.0, 9000.0, 9000.0, 9000.0, 9000.0, 9000.0, 9000.0, 9000.0, 9000.0};
float HighPressureCoeff[10] = {1.8, 1.8, 1.8, 1.8, 1.8, 1.8, 1.8, 1.8, 1.8, 1.8};
//float LowPressureCoeff[10] = {0.8, 0.4, 1.4, 0.4, 1.4, 0.4, 1.5, 0.4, 0.4, 0.4};        //C4 should be smaller

/*Parameter for soft gripper */         //6,7,9 =OK
float LowPressureCoeff[10] = {1.3, 0.9, 1.3,   1.3, 1.3, 1.3,   1.5, 1.3, 0.4, 1.3};        //C4 should be smaller

volatile float KP[10] = {20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0};
volatile float KI[10] = {0.0, 0.0, 0.0, 0.0, 0.0,  0.0, 0.0, 0.0, 0.0, 0.0};  //15 -> 25 -> 10
volatile float KD[10] = {0.0, 0.0, 0.0, 0.0, 0.0,  0.0, 0.0, 0.0, 0.0, 0.0};

//Variables used during calculation
volatile float pressure_err[10] = {0.0, 0.0, 0.0, 0.0, 0.0,  0.0, 0.0, 0.0, 0.0, 0.0};
volatile float pressure_err_last[10] = {0.0, 0.0, 0.0, 0.0, 0.0,  0.0, 0.0, 0.0, 0.0, 0.0};
volatile float pressure_err_integral[10] = {0.0, 0.0, 0.0, 0.0, 0.0,  0.0, 0.0, 0.0, 0.0, 0.0};
volatile float feedbackSpeed[10] = {0.0, 0.0, 0.0, 0.0, 0.0,  0.0, 0.0, 0.0, 0.0, 0.0};

volatile float commandSpeed_pump[2] = {0.0,0.0};
              
volatile float currentPressure[10]={0.0, 0.0, 0.0, 0.0, 0.0,  0.0, 0.0, 0.0, 0.0, 0.0};
volatile float lastPressure[10]={0.0, 0.0, 0.0, 0.0, 0.0,  0.0, 0.0, 0.0, 0.0, 0.0};
volatile float targetPressure[10]={0.0, 0.0, 0.0, 0.0, 0.0,  0.0, 0.0, 0.0, 0.0, 0.0};

volatile float global_Debug_Para = 1.0;

HardwareTimer *MyTim = new HardwareTimer(TIM2);   
int counter_test = 0;

//it will run before main loop for once
void setup() 
{

  ioport_1.begin(); // Begin I2C
   // Set (IO Expander 1) first 14 pins to output
  for (uint8_t i = 0; i < 10; i++){
    ioport_1.pinMode(i, OUTPUT);
  }

  ioport_1.digitalWrite(S00, HIGH);  // Set the SS pin HIGH
  ioport_1.digitalWrite(S01, HIGH);  // Set the SS pin HIGH
  ioport_1.digitalWrite(S02, HIGH);  // Set the SS pin HIGH
  ioport_1.digitalWrite(S03, HIGH);  // Set the SS pin HIGH
  ioport_1.digitalWrite(S04, HIGH);  // Set the SS pin HIGH
  ioport_1.digitalWrite(S05, HIGH);  // Set the SS pin HIGH
  ioport_1.digitalWrite(S06, HIGH);  // Set the SS pin HIGH
  ioport_1.digitalWrite(S07, HIGH);  // Set the SS pin HIGH
  ioport_1.digitalWrite(S08, HIGH);  // Set the SS pin HIGH
  ioport_1.digitalWrite(S09, HIGH);  // Set the SS pin HIGH
  
  
  /*Initialization of pwm generator*/
  pwm0.begin();
  pwm0.setPWMFreq(1000);  // This is the maximum PWM frequency

  pwm1.begin();
  pwm1.setPWMFreq(1000);  // This is the maximum PWM frequency


//  pwmGenerator.setPWM(CHANNEL0_PUMP0_PIN, 0, 0); 
//  pwmGenerator.setPWM(CHANNEL0_PUMP1_PIN, 0, 0); 
//  pwmGenerator.setPWM(CHANNEL0_VALVE0_PIN, 0, 0); 
//  
  //afio_cfg_debug_ports(AFIO_DEBUG_NONE);// Stop the debug function
  // Initialize digital pins as output
//  pinMode(PA1, OUTPUT);

  Serial.setRx(PA10);
  Serial.setTx(PA9);

  SPI.begin();  // Begin SPI hardware
  SPI.setClockDivider(SPI_CLOCK_DIV64);  // Slow down SPI clock
  //SPI.setDataMode(SPI_MODE2);

  // Initialize serial
  Serial.begin(115200); 
  Serial.setTimeout(1000); 

  
  delay(10);
  MyTim->setOverflow(30, HERTZ_FORMAT); // 100 Hz, but only 50Hz is used to pressure PID
  MyTim->attachInterrupt(PressurePID_Ctrller);


  closeValve();
}

//after running the setup() once, this loop() will be ran endlessly
//this is the main loop
//we usually put some communication code here.
int globalChannelNo = 0;
int gameCounter = 0;
void loop() 
{
if (Serial.available() > 0) 
{
  inputString = String(Serial.readStringUntil(10));//search for \n
    if(inputString.endsWith(String('\r')))
    {
      //
      // AT command test
      //
      if (inputString=="AT\r"){
        Serial.println("OK;Successful");
      } 
      //
      // Get board version
      //
      else if(inputString=="AT+VERSION\r"){
        Serial.print("Version 2.0");
        Serial.println("OK");
      } 
      //
      // Get address of this device
      //
      else if(inputString=="AT+ADDRESS\r"){
        Serial.println(BOARD_ADDRESS);
        //Serial.println("OK");
      } 
       else if(inputString=="AT+PRESSURE\r"){
        
        Serial.println(getPressure(S00));
        //Serial.println("OK");
      } 

      //LowPressureCoeff
       else if(inputString.substring(0,3)=="SP="){

        int idx = inputString.indexOf('\r');
        float P = inputString.substring(3,idx).toFloat();
        LowPressureCoeff[7] = P;
        LowPressureCoeff[9] = P;
        Serial.println("Change P to "+String(P));
        
       }
      else if(inputString.substring(0,3)=="SC="){

        int idx = inputString.indexOf('\r');
        globalChannelNo = inputString.substring(3,idx).toInt();
        Serial.println("Global Channel is "+String(globalChannelNo));
      }

      else if(inputString.substring(0,3)=="ST="){

        int idx = inputString.indexOf('\r');
        thresholdP = inputString.substring(3,idx).toFloat();
        Serial.println("Global Threshold is "+String(thresholdP));
      } 
      
      
      else if(inputString.substring(0,2)=="T="){

        int idx = inputString.indexOf('\r');
        float targetP = inputString.substring(2,idx).toFloat();

        //0 2 4 6      8
        if(globalChannelNo%2==0)
        {
           targetPressure[globalChannelNo] = targetP;
           targetPressure[globalChannelNo+1] = targetP;
         }
         //1 3 5 7     9
        else
        {
           targetPressure[globalChannelNo] = targetP;
           targetPressure[globalChannelNo-1] = targetP;
        }
        
        
//        targetPressure[6] = targetP;
//        targetPressure[7] = targetP;

//        targetPressure[1] = targetP;
//        targetPressure[3] = targetP;
//        targetPressure[5] = targetP;
//        targetPressure[7] = targetP;
//        targetPressure[9] = targetP;
        //Serial.println("Global Channel Target Pressure is "+String(targetPressure[globalChannelNo]));
      } 

      else if(inputString.substring(0,4)=="ALLS"){
//       targetPressure[0] = -50;
//      targetPressure[2] = -50;
//      targetPressure[4] = -50;
//      targetPressure[6] = -50;
//      targetPressure[8] = -50;

//      targetPressure[1] = 60;
//      targetPressure[3] = 60;
//      targetPressure[5] = 0;
//      targetPressure[7] = 0;
//      targetPressure[9] = 60;
      
        targetPressure[1] = 50;
        targetPressure[3] = 50;
        targetPressure[5] = 50;
        targetPressure[7] = 50;
        targetPressure[9] = 50;
        
        Serial.println("Target Array Set..");
      } 

       else if(inputString.substring(0,4)=="ALLD"){
      targetPressure[0] = -0.1;
      targetPressure[2] = -0.1;
      targetPressure[4] = -0.1;
      targetPressure[6] = -0.1;
      targetPressure[8] = -0.1;

      targetPressure[1] = -0.1;
      targetPressure[3] = -0.1;
      targetPressure[5] = -0.1;
      targetPressure[7] = -0.1;
      targetPressure[9] = -0.1;
      
        
        Serial.println("Target Array Date Back..");
      } 

      /*"Rock Paper Scissors*/
      else if(inputString.substring(0,4)=="GAME"){
       if(gameCounter%7==0)
       {
          /*Rock vs Rock*/
          targetPressure[0] = -50;
          targetPressure[2] = -50;
          targetPressure[4] = -15;
          targetPressure[6] = -40;
          targetPressure[8] = -50;
    
          targetPressure[1] = 50;
          targetPressure[3] = 50;
          targetPressure[5] = 50;
          targetPressure[7] = 50;
          targetPressure[9] = 50;
       }

       if(gameCounter%7==1)
       {
          
          targetPressure[0] = -50;
          targetPressure[2] = -50;
          targetPressure[4] = 0;
          targetPressure[6] = 0;
          targetPressure[8] = -50;
    
          targetPressure[1] = 0;
          targetPressure[3] = 0;
          targetPressure[5] = 0;
          targetPressure[7] = 0;
          targetPressure[9] = 0;
       }

       if(gameCounter%7==2)
       {
          /*Rock vs Rock*/
          targetPressure[0] = -50;
          targetPressure[2] = -50;
          targetPressure[4] = -15;
          targetPressure[6] = -40;
          targetPressure[8] = -50;
    
          targetPressure[1] = 50;
          targetPressure[3] = 50;
          targetPressure[5] = 50;
          targetPressure[7] = 50;
          targetPressure[9] = 50;
       }

       if(gameCounter%7==3)
       {
         
          targetPressure[0] = -50;
          targetPressure[2] = -50;
          targetPressure[4] = -15;
          targetPressure[6] = -40;
          targetPressure[8] = -50;

          targetPressure[1] = 50;
          targetPressure[3] = 50;
          targetPressure[5] = 0;
          targetPressure[7] = 0;
          targetPressure[9] = 50;
          
       }

       if(gameCounter%7==4)
       {
          /*Rock vs Rock*/
          targetPressure[0] = -50;
          targetPressure[2] = -50;
          targetPressure[4] = -15;
          targetPressure[6] = -40;
          targetPressure[8] = -50;
    
          targetPressure[1] = 50;
          targetPressure[3] = 50;
          targetPressure[5] = 50;
          targetPressure[7] = 50;
          targetPressure[9] = 50;
          
       }

       if(gameCounter%7==5)
       {
          
          targetPressure[0] = 0;
          targetPressure[2] = 0;
          targetPressure[4] = 0;
          targetPressure[6] = 0;
          targetPressure[8] = 0;
    
          targetPressure[1] = 50;
          targetPressure[3] = 50;
          targetPressure[5] = 50;
          targetPressure[7] = 50;
          targetPressure[9] = 50;
          
       }

       if(gameCounter%7==6)
       {
          
          targetPressure[0] = 0;
          targetPressure[2] = 0;
          targetPressure[4] = 0;
          targetPressure[6] = 0;
          targetPressure[8] = 0;

          targetPressure[1] = 0;
          targetPressure[3] = 0;
          targetPressure[5] = 0;
          targetPressure[7] = 0;
          targetPressure[9] = 0;
          
       }

        
        gameCounter = gameCounter + 1;
      }
      //mannequin
      else if(inputString.substring(0,3)=="MAN"){
        if(gameCounter%2==0)
       {
          targetPressure[0] = 5.498634;
          targetPressure[1] = 5.498634;
          
          targetPressure[2] = 0.93600;  
          targetPressure[3] = 0.93600;
          
          targetPressure[4] = 9.945285;
          targetPressure[5] = 9.945285;
          
          targetPressure[6] = 5.921213;
          targetPressure[7] = 5.921213;

          targetPressure[8] = 0.93600;  
          targetPressure[9] = 0.93600;

       }

        if(gameCounter%2==1)
       {
          targetPressure[0] = -0.1;
          targetPressure[1] = -0.1;
          
          targetPressure[2] = -0.1; 
          targetPressure[3] = -0.1;
          
          targetPressure[4] = -0.1;
          targetPressure[5] = -0.1;
          
          targetPressure[6] = -0.1;
          targetPressure[7] = -0.1;

          targetPressure[8] = -0.1; 
          targetPressure[9] = -0.1;
         
       }
        gameCounter = gameCounter + 1;
      }

      /*"Rock Paper Scissors*/
      else if(inputString.substring(0,7)=="GRIPPER"){
       if(gameCounter%5==0)
       {
          targetPressure[0] = 25;
          targetPressure[1] = 28;
          targetPressure[2] = 0.001;
          
          targetPressure[3] = 0.001;
          targetPressure[4] = 29;
          targetPressure[5] = 29;
          
          targetPressure[6] = 25.0;
          targetPressure[7] = 0.001;
          targetPressure[9] = 25.0;
       }

       if(gameCounter%5==1)
       {
          targetPressure[0] = -0.001;
          targetPressure[1] = -0.001;
          targetPressure[2] = -0.2;
          
          targetPressure[3] = -0.001;
          targetPressure[4] = -0.001;
          targetPressure[5] = -0.001;
          
          targetPressure[6] = -0.001;
          targetPressure[7] = -0.001;
          targetPressure[9] = -0.001;
       }
       

       if(gameCounter%5==2)
       {
          targetPressure[0] = -0.001;
          targetPressure[1] = -0.001;
          targetPressure[2] = 25.0;
          
          targetPressure[3] = 29.0;
          targetPressure[4] = -0.001;
          targetPressure[5] = -0.001;
          
          targetPressure[6] = -0.001;
          targetPressure[7] = 25.0;
          targetPressure[9] = -0.001;
       }

       if(gameCounter%5==3)
       {
          targetPressure[0] = 25.0;
          targetPressure[1] = -0.001;
          targetPressure[2] = 25.0;
          
          targetPressure[3] = 29.0;
          targetPressure[4] = 29.0;
          targetPressure[5] = -0.001;
          
          targetPressure[6] = -0.001;
          targetPressure[7] = 25.0;
          targetPressure[9] = 25.0;
          
       }
        if(gameCounter%5==4)
       {
          targetPressure[0] = -0.001;
          targetPressure[1] = -0.001;
          targetPressure[2] = -0.2;
          
          targetPressure[3] = -0.001;
          targetPressure[4] = -0.001;
          targetPressure[5] = -0.001;
          
          targetPressure[6] = -0.001;
          targetPressure[7] = -0.001;
          targetPressure[9] = -0.001;
          
       }

        
        gameCounter = gameCounter + 1;
      }
      
      else if(inputString.substring(0,2)=="M="){

        int idx = inputString.indexOf('\r');
        float targetP = inputString.substring(2,idx).toFloat();
        global_Debug_Para = targetP;
       // Serial.println(String(global_Debug_Para));
      } 
      else if(inputString=="C\r"){

        MyTim->resume();
        Serial.println("Timer Starts...");
      } 

      else if(inputString=="V\r"){

        MyTim->pause();
        for(int i=0;i<10;i++)
        {
          adjustSpeedWithChannel(i,-0.2);
        }
        Serial.println("Timer Pauses...");
        Serial.println("Stop All Motors");
      } 

       else if(inputString=="B\r"){
        Serial.println(String(getPressure(S00)));
        Serial.println("Suck");
      } 
//
      else if(inputString=="H\r"){
        pwm0.setPWM(0, 0, 4000); 
        pwm0.setPWM(1, 0, 4000); 
        //pwmGenerator[0].setPWM(2, 0, 4000); 
        pwm0.setPWM(3, 0, 4000); 
        pwm0.setPWM(4, 0, 4000); 
        Serial.println("Inflation");
      }

      

      else if(inputString=="J\r"){
        pwm0.setPWM(0, 0, 0); 
        pwm0.setPWM(1, 0, 0); 
        pwm0.setPWM(2, 0, 0); 
        pwm0.setPWM(3, 0, 0); 
        pwm0.setPWM(4, 0, 0); 
        Serial.println("LOW");
      }

      else if(inputString=="OO\r"){
        adjustSpeed(0, -4090);
        //pwmGenerator.setPWM(CHANNEL0_VALVE0_PIN, 0, 4095); 
        //valveState[0] = true;

        Serial.println("OO");
        
      }
      else if(inputString=="LL\r"){
        adjustSpeed(0, 4090);
        //pwmGenerator.setPWM(CHANNEL0_VALVE0_PIN, 0, 4095); 
        //valveState[0] = true;

        Serial.println("LL");
        
      }
      else if(inputString=="P\r"){
        
        adjustSpeedWithChannel(globalChannelNo,-0.2);

        Serial.println("Stop" + String(globalChannelNo));
      }
      else if(inputString=="PP\r"){

        for(int i=0;i<10;i++)
        {
          adjustSpeedWithChannel(i,-0.2);
        }
        Serial.println("Stop all");
      }
      else if(inputString=="UU\r"){

        //pwmGenerator2.setPWM(6, 0, 4092); 
         openValve();
          Serial.println("Open Valve");
      }

      else if(inputString=="II\r"){
       
       //pwmGenerator2.setPWM(6, 0, 0); 
        closeValve();
        Serial.println("Close Valve");
      }

      else if(inputString=="CPU\r"){
       
       
        Serial.println(F_CPU);
      }
      
      else{
        Serial.println("Error command");
      }
    }
    else{
      Serial.println("Error end check");
    }
  
 }
 
}


//Timer function
//It will run several milliseconds or several microseconds. (Frequency)
/*PID Controller for Pressure*/
volatile int count = 0;
volatile int counter_pid = 0;
volatile float finitePressure[2] = {0.0,0.0};
volatile float finiteDifferentPressure = 0.0;
float t_interval = 1.0/50.0;
// 100 Hz as 1/100.0 s
//adjustSpeedWithChannel
void PressurePID_Ctrller(void)
{

//    // //every 1/5 
//    // /*Acquire Current Pressure Value*/
//    // currentPressure[0] = getPressure(S00);
//  
//    // //PID with adjustable vel
//    // pressure_err[0] = currentPressure[0] - targetPressure[0];
//    // pressure_err_integral[0] = pressure_err_integral[0] + pressure_err[0];
//  
//    
//    // /*
//     // * Positive Pressure Regulation
//     // */
//    
//    // if(targetPressure[0]>=0)
//    // {
//      // /*Positve Pressure Situation: Deflation*/
//      // if(pressure_err[0]>1.0)
//      // {
//      // //when target pressure is lower than current one (larger than 1 KPa), only switch the air channel to deflation 
//      // //until pressure is below target pressure, then switch back
//      // feedbackSpeed[0] = -1.0;
//      // commandSpeed_pump[0] = 0.0;
//      // commandSpeed_pump[1] = 0.0;
//      // }
//      // /*Positve Pressure Situation: Inflation*/
//      // else
//      // {
//        // if(pressure_err[0]>0)
//        // {
//        // //we just need to wait for system to drop to pressure_err[0]<0
//        // feedbackSpeed[0] = 0.01;
//        // commandSpeed_pump[0] = 0.00;
//        // commandSpeed_pump[1] = 0.0;
//        // }
//        // else
//        // {
//          // if(fabs(pressure_err[0])>5)
//          // {
//            // //NICE
//            // KP[0] = KP_Original[0]*0.12*LowPressureCoeff[0]; 
//          // }
//            
//          // else if(fabs(pressure_err[0])>3)
//          // {
//            // //3->5 0.08
//            // //NICE
//             // KP[0] = KP_Original[0]*0.18*LowPressureCoeff[0]; 
//          // }
//            
//             
//          // else if(fabs(pressure_err[0])>1)
//          // {
//            // //1->3
//            // //PARAMETER IS NICE
//            // KP[0] = KP_Original[0]*0.2*LowPressureCoeff[0]; 
//          // }
//          
//          // else if(fabs(pressure_err[0])>0.5)
//          // {
//            // //0.5->1
//            // //0.4 TOO MUCH 
//            // KP[0] = KP_Original[0]*0.32*LowPressureCoeff[0]; 
//          // }
//        
//          // else
//          // {
//            // //0 -> 0.5
//            // //NICE
//            // KP[0] = KP_Original[0]*0.6*LowPressureCoeff[0]; 
//          // }
//        
//          // // feedbackSpeed[0] = fabs(-1.0*(KP[0]*pressure_err[0]+KI[0]*pressure_err_integral[0]+KD[0]*(pressure_err[0]-pressure_err_last[0])));
//          // feedbackSpeed[0] = fabs(KP[0]*pressure_err[0]);
//    
//          // commandSpeed_pump[0] = feedbackSpeed[0]/1000.0;
//          // commandSpeed_pump[1] = 0.0;
//        
//        // }
//        
//      // }
//    // }
//    
//  
//    // /*
//     // * Negative Pressure Regulation
//     // */
//    // if(targetPressure[0]<0)
//    // {
//      // /*Negative Pressure Situation: Inflation*/
//      // if(pressure_err[0]<-1.0)
//      // {
//      // //when target pressure is higher than current one (larger than 1 KPa), only switch the air channel to inflation 
//      // //until pressure is beyond target pressure, then switch back
//      // feedbackSpeed[0] = 1.0;
//      // commandSpeed_pump[0] = 0.0;   //inflation pump
//      // commandSpeed_pump[1] = 0.0;   //deflation pump
//      // }
//      
//      // //below is normal situation
//      // /*Negative Pressure Situation: Deflation*/
//      // else
//      // {
//        // if(pressure_err[0]<0)
//        // {
//        // //we just need to wait for system to drop to pressure_err[0]<0
//        // feedbackSpeed[0] = -0.01;
//        // commandSpeed_pump[0] = 0.0;
//        // commandSpeed_pump[1] = 0.00;
//        // }
//        // else
//        // {
//        // //increase vaccum ability
//        // //
//        // if(fabs(pressure_err[0])>5)
//        // {
//          // //NICE
//          // KP[0] = KP_Original[0]*0.12*LowPressureCoeff[0]; 
//        // }
//          
//        // else if(fabs(pressure_err[0])>3)
//        // {
//          // //3->5 0.08
//          // //NICE
//           // KP[0] = KP_Original[0]*0.18*LowPressureCoeff[0]; 
//        // }
//          
//           
//        // else if(fabs(pressure_err[0])>1)
//        // {
//          // //1->3
//          // //PARAMETER IS NICE
//          // KP[0] = KP_Original[0]*0.2*LowPressureCoeff[0]; 
//        // }
//        
//        // else if(fabs(pressure_err[0])>0.5)
//        // {
//          // //0.5->1
//          // //0.4 TOO MUCH 
//          // KP[0] = KP_Original[0]*0.32*LowPressureCoeff[0]; 
//        // }
//      
//        // else
//        // {
//          // //0 -> 0.5
//          // //NICE
//          // KP[0] = KP_Original[0]*0.6*LowPressureCoeff[0]; 
//        // }
//        
//        // //feedbackSpeed[0] = -1.0*(KP[0]*pressure_err[0]+KI[0]*pressure_err_integral[0]+KD[0]*(pressure_err[0]-pressure_err_last[0]));
//        // feedbackSpeed[0] = -fabs(1.0*(KP[0]*pressure_err[0]));
//  
//        // commandSpeed_pump[0] = 0.0;
//        // commandSpeed_pump[1] = feedbackSpeed[0]/1000.0;
//        
//        // }
//      // }
//    // }
//    
//        
//    
//    // //apply vel change
//    // //after move to multiple channels, we should finish all feedback speed calculation first, then apply vel change
//    // adjustSpeed(0, feedbackSpeed[0]);
//    
//    // //negative pressure regulation
//    // //adjustSpeedN(0, feedbackSpeed[0]);
//    
//    // //update state
//    // pressure_err_last[0] = pressure_err[0];
//  
//  
//  // //  Serial.println(String(getPressure(S02)));
//  
//    // //this information is justed for visualization
//    // commandSpeed_pump[0] = fabs(commandSpeed_pump[0]*1000.0);
//    // if (commandSpeed_pump[0] >=4095)
//      // commandSpeed_pump[0] = 4095;
//    // commandSpeed_pump[0] = commandSpeed_pump[0]/1000.0;
//      
//    // commandSpeed_pump[1] =fabs(commandSpeed_pump[1]*1000.0);
//    // if (commandSpeed_pump[1] >=4095)
//      // commandSpeed_pump[1] = 4095;
//    // commandSpeed_pump[1] = - commandSpeed_pump[1]/1000.0;
//
//    // finitePressure[0] = finitePressure[1];
//    // finitePressure[1] = currentPressure[0];
//   // finiteDifferentPressure = (finitePressure[1] - finitePressure[0])/t_interval;
//   
//    
//    
//    // //Serial.print("T");
//    // Serial.print(String(currentPressure[0]));
//    // Serial.print(",");
//    // Serial.print(String(targetPressure[0]));
//    // Serial.print(",");
//    // if (valveState[0] ==true)
//      // Serial.print(String(10));
//    // else
//      // Serial.print(String(0));
//    // Serial.print(",");
//    // Serial.print(String(commandSpeed_pump[0]));
//    // Serial.print(",");
//    // Serial.println(String(commandSpeed_pump[1]));
//// //    Serial.print(",");
//// //    Serial.println(String(finiteDifferentPressure));
//// //
//// //      
//  
//   
//
//   
//
//
//   // counter_pid = counter_pid + 1;
//// //    if(counter_pid%500==0)
//// //    {
//// //      Serial.println("10s...."+String(int(counter_pid/500)));
//// //    }
//     
//    // if(counter_pid>=60000)
//    // {
//      // counter_pid = 0;
//    // }
//   // count = count+1;
//    // if(count>=20000)
//    // {
//      // count = 0;
//    // }
//	
//	
//	
//	
//
//	
//
//
//
//	// /*For a specific channelNo*/
//    // /*Acquire Current Pressure Value*/
//    // currentPressure[0] = getPressure(S00);
//  
//    // //PID with adjustable vel
//    // pressure_err[0] = currentPressure[0] - targetPressure[0];
//    // pressure_err_integral[0] = pressure_err_integral[0] + pressure_err[0];
//  
//    
//    // /*
//     // * Positive Pressure Regulation
//     // */
//    
//    // if(targetPressure[0]>=0)
//    // {
//      // /*Positve Pressure Situation: Deflation*/
//      // if(pressure_err[0]>1.0)
//      // {
//      // //when target pressure is lower than current one (larger than 1 KPa), only switch the air channel to deflation 
//      // //until pressure is below target pressure, then switch back
//      // feedbackSpeed[0] = -1.0;
//      // commandSpeed_pump[0] = 0.0;
//      // commandSpeed_pump[1] = 0.0;
//      // }
//      // /*Positve Pressure Situation: Inflation*/
//      // else
//      // {
//        // if(pressure_err[0]>0)
//        // {
//        // //we just need to wait for system to drop to pressure_err[0]<0
//        // feedbackSpeed[0] = 0.01;
//        // commandSpeed_pump[0] = 0.00;
//        // commandSpeed_pump[1] = 0.0;
//        // }
//        // else
//        // {
//          // if(fabs(pressure_err[0])>5)
//          // {
//            // //NICE
//            // KP[0] = KP_Original[0]*0.12*LowPressureCoeff[0]; 
//          // }
//            
//          // else if(fabs(pressure_err[0])>3)
//          // {
//            // //3->5 0.08
//            // //NICE
//             // KP[0] = KP_Original[0]*0.18*LowPressureCoeff[0]; 
//          // }
//            
//             
//          // else if(fabs(pressure_err[0])>1)
//          // {
//            // //1->3
//            // //PARAMETER IS NICE
//            // KP[0] = KP_Original[0]*0.2*LowPressureCoeff[0]; 
//          // }
//          
//          // else if(fabs(pressure_err[0])>0.5)
//          // {
//            // //0.5->1
//            // //0.4 TOO MUCH 
//            // KP[0] = KP_Original[0]*0.32*LowPressureCoeff[0]; 
//          // }
//        
//          // else
//          // {
//            // //0 -> 0.5
//            // //NICE
//            // KP[0] = KP_Original[0]*0.6*LowPressureCoeff[0]; 
//          // }
//        
//          // // feedbackSpeed[0] = fabs(-1.0*(KP[0]*pressure_err[0]+KI[0]*pressure_err_integral[0]+KD[0]*(pressure_err[0]-pressure_err_last[0])));
//          // feedbackSpeed[0] = fabs(KP[0]*pressure_err[0]);
//    
//          // commandSpeed_pump[0] = feedbackSpeed[0]/1000.0;
//          // commandSpeed_pump[1] = 0.0;
//        
//        // }
//        
//      // }
//    // }
//    
//  
//    // /*
//     // * Negative Pressure Regulation
//     // */
//    // if(targetPressure[0]<0)
//    // {
//      // /*Negative Pressure Situation: Inflation*/
//      // if(pressure_err[0]<-1.0)
//      // {
//      // //when target pressure is higher than current one (larger than 1 KPa), only switch the air channel to inflation 
//      // //until pressure is beyond target pressure, then switch back
//      // feedbackSpeed[0] = 1.0;
//      // commandSpeed_pump[0] = 0.0;   //inflation pump
//      // commandSpeed_pump[1] = 0.0;   //deflation pump
//      // }
//      
//      // //below is normal situation
//      // /*Negative Pressure Situation: Deflation*/
//      // else
//      // {
//        // if(pressure_err[0]<0)
//        // {
//        // //we just need to wait for system to drop to pressure_err[0]<0
//        // feedbackSpeed[0] = -0.01;
//        // commandSpeed_pump[0] = 0.0;
//        // commandSpeed_pump[1] = 0.00;
//        // }
//        // else
//        // {
//        // //increase vaccum ability
//        // //
//        // if(fabs(pressure_err[0])>5)
//        // {
//          // //NICE
//          // KP[0] = KP_Original[0]*0.12*LowPressureCoeff[0]; 
//        // }
//          
//        // else if(fabs(pressure_err[0])>3)
//        // {
//          // //3->5 0.08
//          // //NICE
//           // KP[0] = KP_Original[0]*0.18*LowPressureCoeff[0]; 
//        // }
//          
//           
//        // else if(fabs(pressure_err[0])>1)
//        // {
//          // //1->3
//          // //PARAMETER IS NICE
//          // KP[0] = KP_Original[0]*0.2*LowPressureCoeff[0]; 
//        // }
//        
//        // else if(fabs(pressure_err[0])>0.5)
//        // {
//          // //0.5->1
//          // //0.4 TOO MUCH 
//          // KP[0] = KP_Original[0]*0.32*LowPressureCoeff[0]; 
//        // }
//      
//        // else
//        // {
//          // //0 -> 0.5
//          // //NICE
//          // KP[0] = KP_Original[0]*0.6*LowPressureCoeff[0]; 
//        // }
//        
//        // //feedbackSpeed[0] = -1.0*(KP[0]*pressure_err[0]+KI[0]*pressure_err_integral[0]+KD[0]*(pressure_err[0]-pressure_err_last[0]));
//        // feedbackSpeed[0] = -fabs(1.0*(KP[0]*pressure_err[0]));
//  
//        // commandSpeed_pump[0] = 0.0;
//        // commandSpeed_pump[1] = feedbackSpeed[0]/1000.0;
//        
//        // }
//      // }
//    // }
//    
//        
//    
//    // //apply vel change
//    // //after move to multiple channels, we should finish all feedback speed calculation first, then apply vel change
//    // adjustSpeed(0, feedbackSpeed[0]);
//    
//    // //negative pressure regulation
//    // //adjustSpeedN(0, feedbackSpeed[0]);
//    
//    // //update state
//    // pressure_err_last[0] = pressure_err[0];
//  
//  
//  // //  Serial.println(String(getPressure(S02)));
//  
//    // //this information is justed for visualization
//    // commandSpeed_pump[0] = fabs(commandSpeed_pump[0]*1000.0);
//    // if (commandSpeed_pump[0] >=4095)
//      // commandSpeed_pump[0] = 4095;
//    // commandSpeed_pump[0] = commandSpeed_pump[0]/1000.0;
//      
//    // commandSpeed_pump[1] =fabs(commandSpeed_pump[1]*1000.0);
//    // if (commandSpeed_pump[1] >=4095)
//      // commandSpeed_pump[1] = 4095;
//    // commandSpeed_pump[1] = - commandSpeed_pump[1]/1000.0;
//
//    // finitePressure[0] = finitePressure[1];
//    // finitePressure[1] = currentPressure[0];
//   // finiteDifferentPressure = (finitePressure[1] - finitePressure[0])/t_interval;
//
//
//   /*Control all ten channels at the same time*/
////    for(int i=0;i<4;i++)
////    {
//////      if(i==5)
//////        continue;
////      SubPIDctrller(i*2);
////      //delayMicroseconds(300);
////    }
//    // //Serial.print("T");
//  
//   
//
////    if(counter_pid%10==0)
////    {
////       String output = "";
////      for(int i=0;i<10;i++)
////      {
////        //String output1 = String(i)+":"+String(round(currentPressure[i]*100))+"->"+String(round(targetPressure[i]*100))+";";
////        String output1 = String(i)+":"+String(currentPressure[i])+"->"+String(targetPressure[i])+";";
////        output = output + output1;
////       
////      }
////      Serial.println(output);
////    }
//      
//    /*Control all ten channels at the same time*/
//
//
//  
//
//    
//      
//    /*Control all ten channels at the same time*/
//    /*Control only one channel at the same time*/
//   // SubPIDctrller(0);
////      SubPIDctrller(globalChannelNo);
////      //globalChannelNo
////  
////      // //Serial.print("T");
////      String output1 = String(globalChannelNo)+":"+String(currentPressure[globalChannelNo])+"->"+String(targetPressure[globalChannelNo])+";";
////      Serial.println(output1);
//
//    /*Control only one channel at the same time*/
//
////
////    currentPressure[2] = getPressure(2);
////    Serial.println(currentPressure[2]);
//
////    Serial.print("T");
////    Serial.print(String(currentPressure[globalChannelNo]));
////    Serial.print(",");
////    Serial.print(String(targetPressure[globalChannelNo]));
////    Serial.print(",");
////     if (valveState[0] ==true)
////       Serial.print(String(10));
////     else
////       Serial.print(String(0));
////     Serial.print(",");
////     Serial.print(String(commandSpeed_pump[0]));
////     Serial.print(",");
////     Serial.println(String(commandSpeed_pump[1])+";");
////
//
////  //output four channel pressure values.
      
//      SubPIDctrller(0);
//      SubPIDctrller(1);
      SubPIDctrller(2);
      SubPIDctrller(3);
//      SubPIDctrller(4);
//      SubPIDctrller(5);
//      SubPIDctrller(6);
//      SubPIDctrller(7);
//
//
//     
//      SubPIDctrller(8);
//      SubPIDctrller(9);

 //   SubPIDctrller(6);
    //Serial.print("T");
//    Serial.print(String(currentPressure[6]));
//    Serial.print(",");
//    Serial.print(String(targetPressure[6]));
//    Serial.print(",");
//     if (valveState[0] ==true)
//       Serial.print(String(10));
//     else
//       Serial.print(String(0));
//     Serial.print(",");
//     Serial.print(String(commandSpeed_pump[0]));
//     Serial.print(",");
//     Serial.println(String(commandSpeed_pump[1])+",");

//only update pressure
//  for(int i=0;i<10;i++)
//  {
//    float tempP = getPressure(SSPIN[i]);
////    if(fabs(tempP)>100)
////    {
////      return;
////    }
//    currentPressure[SSPIN[i]] = tempP;
//    delayMicroseconds(200);
//  }
//  
//
    //"T"+
   Serial.println("T"+String(currentPressure[0])+","+String(currentPressure[1])+","+String(currentPressure[2])+","+String(currentPressure[3])+","+String(currentPressure[4])+","+String(currentPressure[5])+","+String(currentPressure[6])+","+String(currentPressure[7])+","+String(currentPressure[8])+","+String(currentPressure[9])+";");
//  //  Serial.println(String(currentPressure[0])+","+String(currentPressure[1])+","+String(currentPressure[2])+","+String(currentPressure[3])+","+String(currentPressure[4]));
//Serial.println("T"+String(currentPressure[0])+","+String(currentPressure[1])+","+String(currentPressure[2])+","+String(currentPressure[3])+","+String(currentPressure[4])+","+String(currentPressure[5])+","+String(currentPressure[6])+","+String(currentPressure[7])+","+String(currentPressure[9])+";"+String(targetPressure[0])+","+String(targetPressure[1])+","+String(targetPressure[2])+","+String(targetPressure[3])+","+String(targetPressure[4])+","+String(targetPressure[5])+","+String(targetPressure[6])+","+String(targetPressure[7])+","+String(targetPressure[9])+",");



////    
//    
//    //Serial.println(String(currentPressure[0])+","+String(targetPressure[0]));
//    
////  Serial.println("T"+String(currentPressure[0])+","+String(targetPressure[0])+","+String(currentPressure[2])+","+String(targetPressure[2])+","+String(currentPressure[4])+","+String(targetPressure[4])+","+String(currentPressure[6])+","+String(targetPressure[6])+";");
////  
//
//	counter_pid = counter_pid + 1;
//    if(counter_pid>=60000)
//		counter_pid = 0;
//	count = count+1;
//	if(count>=20000)
//		count = 0;
//
//
////  float value[10];
////  for(int i=0;i<10;i++)
////  {
////    value[i] = getPressure(i);
////    delayMicroseconds(2000);
////  }
////
////  for(int i=0;i<10;i++)
////  {
////    Serial.print(String(i)+":"+value[i]+"; ");  
////  }
////  
////  Serial.println("...");
//    
  
}


/*Channel No is from 0 to 9*/

void SubPIDctrller(int ChannelNo)
{
	/*For a specific channelNo*/
    /*Acquire Current Pressure Value*/
    float tempP = -1;
    if(ChannelNo%2==0)
    {
       //0 2 4 6 
       tempP = getPressure(SSPIN[ChannelNo]);
    }
    else
    {
      // 1 3 5 7
      tempP = currentPressure[ChannelNo-1];
    }
    
    
    if(fabs(tempP)>100)
    {
      return;
    }
    //delayMicroseconds(200);
    currentPressure[ChannelNo] = tempP;
    //PID with adjustable vel
    pressure_err[ChannelNo] = currentPressure[ChannelNo] - targetPressure[ChannelNo];
    pressure_err_integral[ChannelNo] = pressure_err_integral[ChannelNo] + pressure_err[ChannelNo];
  
    
    /*
     * Positive Pressure Regulation
     */
    
    if(targetPressure[ChannelNo]>=0)
    {
      /*Positve Pressure Situation: Deflation*/
      if(pressure_err[ChannelNo]>thresholdP)
      {
      //when target pressure is lower than current one (larger than 1 KPa), only switch the air channel to deflation 
      //until pressure is below target pressure, then switch back
      feedbackSpeed[ChannelNo] = -1.0;
      // commandSpeed_pump[0] = 0.0;
      // commandSpeed_pump[1] = 0.0;
      }
      /*Positve Pressure Situation: Inflation*/
      else
      {
        if(pressure_err[ChannelNo]>0)
        {
        //we just need to wait for system to drop to pressure_err[0]<0
        feedbackSpeed[ChannelNo] = 0.01;
        // commandSpeed_pump[0] = 0.00;
        // commandSpeed_pump[1] = 0.0;
        }
        else
        {
          if(fabs(pressure_err[ChannelNo])>5)
          {
            //NICE
            KP[ChannelNo] = KP_Original[ChannelNo]*0.12*LowPressureCoeff[ChannelNo]; 
          }
            
          else if(fabs(pressure_err[ChannelNo])>3)
          {
            //3->5 0.08
            //NICE
             KP[ChannelNo] = KP_Original[ChannelNo]*0.18*LowPressureCoeff[ChannelNo]; 
          }
            
             
          else if(fabs(pressure_err[ChannelNo])>1)
          {
            //1->3
            //PARAMETER IS NICE
            KP[ChannelNo] = KP_Original[ChannelNo]*0.2*LowPressureCoeff[ChannelNo]; 
          }
          
          else if(fabs(pressure_err[ChannelNo])>0.5)
          {
            //0.5->1
            //0.4 TOO MUCH 
            KP[ChannelNo] = KP_Original[ChannelNo]*0.32*LowPressureCoeff[ChannelNo]; 
          }
        
          else
          {
            //0 -> 0.5
            //NICE
            KP[ChannelNo] = KP_Original[ChannelNo]*0.6*LowPressureCoeff[ChannelNo]; 
          }
        
          // feedbackSpeed[0] = fabs(-1.0*(KP[0]*pressure_err[0]+KI[0]*pressure_err_integral[0]+KD[0]*(pressure_err[0]-pressure_err_last[0])));
          feedbackSpeed[ChannelNo] = fabs(KP[ChannelNo]*pressure_err[ChannelNo]);
    
          // commandSpeed_pump[0] = feedbackSpeed[0]/1000.0;
          // commandSpeed_pump[1] = 0.0;
        
        }
        
      }
    }
    
  
    /*
     * Negative Pressure Regulation
     */
    if(targetPressure[ChannelNo]<0)
    {
      /*Negative Pressure Situation: Inflation*/
      if(pressure_err[ChannelNo]<-thresholdP)
      {
      //when target pressure is higher than current one (larger than 1 KPa), only switch the air channel to inflation 
      //until pressure is beyond target pressure, then switch back
      feedbackSpeed[ChannelNo] = 1.0;
      // commandSpeed_pump[0] = 0.0;   //inflation pump
      // commandSpeed_pump[1] = 0.0;   //deflation pump
      }
      
      //below is normal situation
      /*Negative Pressure Situation: Deflation*/
      else
      {
        if(pressure_err[ChannelNo]<0)
        {
        //we just need to wait for system to drop to pressure_err[0]<0
        feedbackSpeed[ChannelNo] = -0.01;
        // commandSpeed_pump[0] = 0.0;
        // commandSpeed_pump[1] = 0.00;
        }
        else
        {
        //increase vaccum ability
        //
        if(fabs(pressure_err[ChannelNo])>5)
        {
          //NICE
          KP[ChannelNo] = KP_Original[ChannelNo]*0.12*LowPressureCoeff[ChannelNo]; 
        }
          
        else if(fabs(pressure_err[ChannelNo])>3)
        {
          //3->5 0.08
          //NICE
           KP[ChannelNo] = KP_Original[ChannelNo]*0.18*LowPressureCoeff[ChannelNo]; 
        }
          
           
        else if(fabs(pressure_err[ChannelNo])>1)
        {
          //1->3
          //PARAMETER IS NICE
          KP[ChannelNo] = KP_Original[ChannelNo]*0.2*LowPressureCoeff[ChannelNo]; 
        }
        
        else if(fabs(pressure_err[ChannelNo])>0.5)
        {
          //0.5->1
          //0.4 TOO MUCH 
          KP[ChannelNo] = KP_Original[ChannelNo]*0.32*LowPressureCoeff[ChannelNo]; 
        }
      
        else
        {
          //0 -> 0.5
          //NICE
          KP[ChannelNo] = KP_Original[ChannelNo]*0.6*LowPressureCoeff[ChannelNo]; 
        }
        feedbackSpeed[ChannelNo] = -fabs(1.0*(KP[ChannelNo]*pressure_err[ChannelNo]));
  
        // commandSpeed_pump[0] = 0.0;
        // commandSpeed_pump[1] = feedbackSpeed[0]/1000.0;
        
        }
      }
    }
    
        
    
    //apply vel change
    //after move to multiple channels, we should finish all feedback speed calculation first, then apply vel change
    adjustSpeedWithChannel(ChannelNo, feedbackSpeed[ChannelNo]);
    pressure_err_last[ChannelNo] = pressure_err[ChannelNo];
  
  
  //  Serial.println(String(getPressure(S02)));
  
    //this information is justed for visualization
    // commandSpeed_pump[0] = fabs(commandSpeed_pump[0]*1000.0);
    // if (commandSpeed_pump[0] >=4095)
      // commandSpeed_pump[0] = 4095;
    // commandSpeed_pump[0] = commandSpeed_pump[0]/1000.0;
      
    // commandSpeed_pump[1] =fabs(commandSpeed_pump[1]*1000.0);
    // if (commandSpeed_pump[1] >=4095)
      // commandSpeed_pump[1] = 4095;
    // commandSpeed_pump[1] = - commandSpeed_pump[1]/1000.0;

    // finitePressure[0] = finitePressure[1];
    // finitePressure[1] = currentPressure[0];
   // finiteDifferentPressure = (finitePressure[1] - finitePressure[0])/t_interval;
}

/*
  Functions to adjust flow rate for both positive and negative pressure regulation
  @channelNo: from 0 to 9 representing each air channel
  @flowRate: value is from -4095 to 4095 -> it can be both positive or negative
*/

void adjustSpeed(int channelNo, float flowRate)
{
//  if(flowRate<0)
//    flowRate=0.1;
  
  if(flowRate<=0)
  {
    //reduce pressure
    //if the valve is not ready, make it ready
    if(valveState[0]==true)
    {
      closeValve();
      
    }
    //1.5 -> 1.3 -> 1.05
    flowRate = fabs(flowRate)*1.0;        //adjust coeff to let two pumps to be equal
    if(flowRate>=4095)
      flowRate = 4095;
    pwm0.setPWM(10, 0, fabs(flowRate));  
    
  }
  else
  {
    //add pressure
    //if the valve is not ready, make it ready
    if(valveState[0]==false)
    {
      openValve();
    }
     flowRate = fabs(flowRate)*1.0;
    if(flowRate>=4095)
      flowRate = 4095;
    pwm0.setPWM(0, 0, fabs(flowRate));  
  }
    
}

void adjustSpeedWithChannel(int channelNo, float flowRate)
{
//  if(flowRate<0)
//    flowRate=0.1;
  
  if(flowRate<=0)
  {
    //reduce pressure
    //if the valve is not ready, make it ready
    if(valveState[channelNo]==true)
    {
      closeValveWithChannel(channelNo);
    }
    //1.5 -> 1.3 -> 1.05
    flowRate = fabs(flowRate)*1.0;        //adjust coeff to let two pumps to be equal
    if(flowRate>=4095)
      flowRate = 4095;
  
  
	//set deflation pump speed
	if(AirChannelDeflationPumpPwmChip[channelNo]==0)
		pwm0.setPWM(AirChannelDeflationPumpPwmPort[channelNo],0,fabs(flowRate));		//set deflation pump flow rate to be zero
	else
		pwm1.setPWM(AirChannelDeflationPumpPwmPort[channelNo],0,fabs(flowRate));		//set deflation pump flow rate to be zero
    
  }
  else
  {
    //add pressure
    //if the valve is not ready, make it ready
    if(valveState[channelNo]==false)
    {
      openValveWithChannel(channelNo);
    }
	
     flowRate = fabs(flowRate)*1.0;
	 
    if(flowRate>=4095)
      flowRate = 4095;
  
  
	if(AirChannelInflationPumpPwmChip[channelNo]==0)
		pwm0.setPWM(AirChannelInflationPumpPwmPort[channelNo],0,flowRate);		//set inflation pump flow rate to be zero
	else
		pwm1.setPWM(AirChannelInflationPumpPwmPort[channelNo],0,flowRate);		//set inflation pump flow rate to be zero
  }
    
}



/*Functions to open and close valve*/
//adjust valve position before adding pressure
void openValve(void)
{
  //pwmGenerator2

  pwm0.setPWM(0, 0, 0);
  pwm0.setPWM(10, 0, 0);
  delayMicroseconds(3000);    //1ms 

  pwm1.setPWM(4, 0, 4095); 
  //pwmGenerator2.setPWM(CHANNEL0_VALVE0_PIN, 0, 4095);
  //delay(5);
  delayMicroseconds(8000);    //3000MS when current is 20kPa, can not resist
  valveState[0] = true;
}

//adjust valve position before reducing pressure

void openValveWithChannel(int channelNo)
{
  
  /*first, set two air pump velocity to be zero.*/
  if(AirChannelInflationPumpPwmChip[channelNo]==0)
	pwm0.setPWM(AirChannelInflationPumpPwmPort[channelNo],0,0);		//set inflation pump flow rate to be zero
  else
	pwm1.setPWM(AirChannelInflationPumpPwmPort[channelNo],0,0);		//set inflation pump flow rate to be zero
  
  if(AirChannelDeflationPumpPwmChip[channelNo]==0)
	pwm0.setPWM(AirChannelDeflationPumpPwmPort[channelNo],0,0);		//set deflation pump flow rate to be zero
  else
	pwm1.setPWM(AirChannelDeflationPumpPwmPort[channelNo],0,0);		//set deflation pump flow rate to be zero
  
  /*second, wait for pump to response, this value can be shorter*/
  delayMicroseconds(1000);    //1ms 
  
  /*third, switch the valve state*/
  pwm1.setPWM(AirChannelValvePwmPort[channelNo],0,4095);
  
  delayMicroseconds(3000);    //3000MS when current is 20kPa, can not resist
  valveState[channelNo] = true;
	
}

void closeValveWithChannel(int channelNo)
{
  
  /*first, set two air pump velocity to be zero.*/
  if(AirChannelInflationPumpPwmChip[channelNo]==0)
	pwm0.setPWM(AirChannelInflationPumpPwmPort[channelNo],0,0);		//set inflation pump flow rate to be zero
  else
	pwm1.setPWM(AirChannelInflationPumpPwmPort[channelNo],0,0);		//set inflation pump flow rate to be zero
  
  if(AirChannelDeflationPumpPwmChip[channelNo]==0)
	pwm0.setPWM(AirChannelDeflationPumpPwmPort[channelNo],0,0);		//set deflation pump flow rate to be zero
  else
	pwm1.setPWM(AirChannelDeflationPumpPwmPort[channelNo],0,0);		//set deflation pump flow rate to be zero
  
  /*second, wait for pump to response, this value can be shorter*/
  delayMicroseconds(1000);    //1ms 
  
  /*third, switch the valve state*/
  pwm1.setPWM(AirChannelValvePwmPort[channelNo],0,0);
  
  delayMicroseconds(3000);    //3000MS when current is 20kPa, can not resist
  valveState[channelNo] = false;
	
}

void closeValve(void)
{
  pwm0.setPWM(0, 0, 0);
  pwm0.setPWM(10, 0, 0);
  delayMicroseconds(2000);    //1ms

  pwm1.setPWM(4,0, 0); 
  //pwmGenerator2.setPWM(CHANNEL0_VALVE0_PIN, 0, 0);
  //delay(5);
  delayMicroseconds(5000);    //3000MS when current is 20kPa, can not resist
                              //5000MS maybe not enough
  valveState[0] = false;
}

/*
  Functions to get pressure value
  @pinSS: Sensor Selection Pin
  
*/


float getPressure(uint8_t pinSS)
{
  ABP_read(pinSS);
  char hexString[4];
  for (int i=0; i<2; i++)
  {
     char format[2];
     sprintf(format, "%02X", ABP_data[i]);
     hexString[i*2] = format[0];
     hexString[i*2+1] = format[1];
  }
  int val = StrToHex(hexString);
  
  float percentage = (float)val/16383.0*100.0;
  float pressurePSI_fromLowerBound;
//  if(pinSS%2==0)
//    pressurePSI_fromLowerBound= (percentage-10.0)/80.0*30.0-15.0;
//  else
//    pressurePSI_fromLowerBound= (percentage-10.0)/80.0*15.0;
//  

  if(pinSS==1)
    pressurePSI_fromLowerBound= (percentage-10.0)/80.0*15.0;
  else
    pressurePSI_fromLowerBound= (percentage-10.0)/80.0*30.0-15.0;

  float pressureKPa = pressurePSI_fromLowerBound*0.006895*1000;
  
  return pressureKPa;


}

int StrToHex(char str[])
{
  return (int) strtol(str, 0, 16);
}

// This function is used to read ABP pressure sensor. 
// It needs to select which sensor you want to read
void ABP_read(uint8_t ABPss)
{
  SPI.setDataMode(SPI_MODE0);
  ioport_1.digitalWrite(ABPss, LOW);
  delayMicroseconds(100);

  for (int i=0; i<2; i++)
  {
    ABP_data[i] = SPI.transfer(0);
  }
  ioport_1.digitalWrite(ABPss, HIGH);
  delayMicroseconds(100);

  
}
